/**
 * Titanium SDK
 * Copyright TiDev, Inc. 04/07/2022-Present
 * Licensed under the terms of the Apache Public License
 * Please see the LICENSE included with this distribution for details.
 */
package org.appcelerator.titanium.view;

import java.io.IOException;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;

import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.drawable.BitmapDrawable;
import android.graphics.drawable.ColorDrawable;
import android.graphics.drawable.Drawable;
import android.graphics.drawable.PaintDrawable;
import android.graphics.drawable.StateListDrawable;
import android.util.AttributeSet;

public class TiBackgroundDrawable extends StateListDrawable
{
	private Drawable background;
	private RectF innerRect;
	private static final int NOT_SET = -1;
	private int alpha = NOT_SET;

	public TiBackgroundDrawable()
	{
		background = new ColorDrawable(Color.TRANSPARENT);
		innerRect = new RectF();
	}

	@Override
	public void draw(Canvas canvas)
	{
		if (background != null) {
			background.setBounds((int) innerRect.left, (int) innerRect.top, (int) innerRect.right,
								 (int) innerRect.bottom);
		}

		canvas.save();

		if (background != null) {
			if (alpha > NOT_SET) {
				background.setAlpha(alpha);
			}
			background.draw(canvas);
		}

		canvas.restore();
	}

	@Override
	protected void onBoundsChange(Rect bounds)
	{
		super.onBoundsChange(bounds);
		innerRect.set(bounds);
	}

	@Override
	protected boolean onStateChange(int[] stateSet)
	{
		boolean changed = super.onStateChange(stateSet);
		changed = setState(stateSet);
		boolean drawableChanged = false;
		if (background != null) {
			//			Log.e("TiBackground", "background="+background.getClass().getSimpleName()+",state.len="+stateSet.length);
			//			for (int i = 0; i < stateSet.length; i++) {
			//				Log.e("TiBackground", "    state[" + i + "]=" + stateSet[i]);
			//			}
			drawableChanged = background.setState(stateSet);
			if (drawableChanged) {
				invalidateSelf();
			}
		}

		return changed || drawableChanged;
	}

	@Override
	public void addState(int[] stateSet, Drawable drawable)
	{
		if (background instanceof StateListDrawable) {
			((StateListDrawable) background).addState(stateSet, drawable);
		}
	}

	@Override
	protected boolean onLevelChange(int level)
	{
		boolean changed = super.onLevelChange(level);
		boolean backgroundChanged = false;
		if (background instanceof StateListDrawable) {
			backgroundChanged = ((StateListDrawable) background).setLevel(level);
		}
		return changed || backgroundChanged;
	}

	@Override
	public void invalidateSelf()
	{
		super.invalidateSelf();
		if (background instanceof StateListDrawable) {
			((StateListDrawable) background).invalidateSelf();
		}
	}

	@Override
	public void invalidateDrawable(Drawable who)
	{
		super.invalidateDrawable(who);
		if (background instanceof StateListDrawable) {
			((StateListDrawable) background).invalidateDrawable(who);
		}
	}

	@Override
	public void inflate(Resources r, XmlPullParser parser, AttributeSet attrs)
		throws XmlPullParserException, IOException
	{
		super.inflate(r, parser, attrs);
		if (background != null) {
			background.inflate(r, parser, attrs);
		}
	}

	public void releaseDelegate()
	{
		if (background != null) {
			if (background instanceof BitmapDrawable) {
				((BitmapDrawable) background).getBitmap().recycle();
			}
			background.setCallback(null);
			background = null;
		}
	}

	public void setBackgroundColor(int backgroundColor)
	{
		//this.background = new ColorDrawable(backgroundColor);
		releaseDelegate();
		this.background = new PaintDrawable(backgroundColor);
	}

	@SuppressWarnings("deprecation")
	public void setBackgroundImage(Bitmap backgroundImage)
	{
		releaseDelegate();
		this.background = new BitmapDrawable(backgroundImage);
	}

	public void setBackgroundDrawable(Drawable drawable)
	{
		releaseDelegate();
		this.background = drawable;
		onStateChange(getState());
	}

	@Override
	public void setAlpha(int alpha)
	{
		super.setAlpha(alpha);
		this.alpha = alpha;
	}

	public Drawable getBackground()
	{
		return background;
	}

	@Override
	public Drawable getCurrent()
	{
		if (background != null) {
			return background.getCurrent();
		}
		return super.getCurrent();
	}
}
